from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional, Protocol

import questionary
import typer
from rich.console import Console

from .issue_service import Issue

console = Console()


class IssuePresenter(Protocol):
    def select_issue_dialog(self, issues: Sequence[Issue]) -> Optional[Issue]:
        ...


@dataclass(frozen=True)
class DefaultIssuePresenter(IssuePresenter):
    manual_prompt: str = "Manual"
    refresh_prompt: str = "Refresh..."

    def _show_entry_dialog(self) -> str:
        return typer.prompt("Title")

    def _show_selection_dialog(self, issues: Sequence[Issue]) -> str:
        issue_titles = [issue.title for issue in issues]
        return questionary.select(
            "What's next? 🚀",
            choices=[self.refresh_prompt, self.manual_prompt, *issue_titles],
            default=issue_titles[0],
        ).ask()

    def select_issue_dialog(self, issues: Sequence[Issue]) -> Optional[Issue]:
        selected_issue_title = self._show_selection_dialog(issues=issues)

        if selected_issue_title == self.refresh_prompt:
            return None

        if selected_issue_title == self.manual_prompt:
            selected_issue_title = self._show_entry_dialog()
            return Issue(entity_id=None, title=selected_issue_title)

        return next(issue for issue in issues if issue.title == selected_issue_title)



import re
from dataclasses import dataclass
from typing import Protocol

from .git_utils import StagingMigrater
from .issue_service import Issue
from .subprocess_utils import interactive_cmd


@dataclass(frozen=True)
class ParsedIssue:
    prefix: str
    description: str


def sanitise_text_for_git(input_string: str) -> str:
    char2replacement = {
        " ": "-",
        ":": "/",
        ",": "",
        "'": "",
        '"': "",
        "(": "",
        ")": "",
        "[": "",
        "": "",
        "`": "",
        ">": "",
        "<": "",
        "=": "",
    }

    for character, replacement in char2replacement.items():
        input_string = input_string.replace(character, replacement)

    return input_string.replace("--", "-")


def parse_issue_title(issue_title: str) -> ParsedIssue:
    # Get all string between start and first ":"
    prefix = re.findall(r"^(.*?):", issue_title)[0]
    description = re.findall(r": (.*)$", issue_title)[0]

    return ParsedIssue(
        prefix=sanitise_text_for_git(input_string=prefix),
        description=sanitise_text_for_git(input_string=description),
    )


class Queuer(Protocol):
    def create_queue_from_trunk(self, issue: Issue):
        ...

    def add_to_beginning_of_queue(self, issue: Issue):
        ...

    def add_to_end_of_queue(self, issue: Issue):
        ...

    def submit_queue(self, automerge: bool):
        ...

    def status(self):
        ...


class Graphite(Queuer):
    def _sync(self):
        interactive_cmd("gt sync --force")

    def create_queue_from_trunk(self, issue: Issue):
        self._sync()
        interactive_cmd("git checkout main")
        interactive_cmd("git pull")
        self.add_to_end_of_queue(issue)

    def add_to_beginning_of_queue(self, issue: Issue):
        self._sync()
        first_commit_str = self._get_first_commit_str(issue)
        branch_title = self._get_branch_title(issue=issue)

        with StagingMigrater():
            interactive_cmd("gt bottom")

        interactive_cmd(f'gt create {branch_title} --all --insert -m "{first_commit_str}"')
        interactive_cmd(f'git commit --allow-empty -m "{first_commit_str}"')

    def add_to_end_of_queue(self, issue: Issue):
        self._sync()
        first_commit_str = self._get_first_commit_str(issue)
        branch_title = self._get_branch_title(issue=issue)
        interactive_cmd(f'gt create {branch_title} --all -m "{first_commit_str}"')
        interactive_cmd(f'git commit --allow-empty -m "{first_commit_str}"')

    def _get_first_commit_str(self, issue: Issue) -> str:
        first_commit_str = f"{issue.title}"
        if issue.entity_id is not None:
            first_commit_str += f"\n\nFixes #{issue.entity_id}"
        return first_commit_str

    def _get_branch_title(self, issue: Issue) -> str:
        parsed_issue = parse_issue_title(issue.title)
        entity_id_section = "" if issue.entity_id is None else f"/{issue.entity_id}"
        return f"{parsed_issue.prefix}{entity_id_section}/{parsed_issue.description}"

    def submit_queue(self, automerge: bool):
        self._sync()
        submit_command = "gt submit --no-edit --publish"

        if automerge:
            submit_command += " --merge-when-ready"

        interactive_cmd(submit_command)

    def status(self):
        interactive_cmd("gt log short --reverse")



from .issue_presenter import DefaultIssuePresenter, IssuePresenter
from .issue_service import GithubIssueService, IssueService
from .queuer import Graphite, Queuer

issue_services: dict[str, type[IssueService]] = {"Github": GithubIssueService}
stackers: dict[str, type[Queuer]] = {"Graphite": Graphite}
presenters: dict[str, type[IssuePresenter]] = {"Default": DefaultIssuePresenter}
